"""
django-pg-bulk-update support.
"""

from django.db.models.sql import Query

from .compatibility import django_pg_bulk_update_available
from .fields import HllField
from .values import HllEmpty, HllValue, HllCombinedExpression

# As django-pg-bulk-update library is not required, import only if it exists
if django_pg_bulk_update_available():
    from django_pg_bulk_update.set_functions import ConcatSetFunction
    from django_pg_bulk_update.compatibility import get_field_db_type
else:
    class ConcatSetFunction:
        pass


    def get_field_db_type(field, conn):
        raise NotImplementedError


class HllConcatFunction(ConcatSetFunction):
    names = {'hll_concat'}
    supported_field_classes = {'HllField'}

    def _parse_null_default(self, field, connection, **kwargs):
        kwargs['null_default'] = kwargs.get('null_default', HllEmpty())
        return super(HllConcatFunction, self)._parse_null_default(field, connection, **kwargs)

    def format_field_value(self, field, val, connection, cast_type=False, **kwargs):
        if not isinstance(field, HllField):
            return super(HllConcatFunction, self).format_field_value(field, val, connection, cast_type=cast_type,
                                                                     **kwargs)

        if not isinstance(val, (HllValue, HllCombinedExpression)):
            raise ValueError('val should be HllValue instance')

        compiler = Query(field.model).get_compiler(connection=connection)
        sql, params = val.as_sql(compiler, connection)

        if cast_type:
            sql = 'CAST(%s AS %s)' % (sql, get_field_db_type(field, connection))

        return sql, tuple(params)



from django.db.models import Aggregate, IntegerField

from django_pg_hll import HllField


class Cardinality(Aggregate):
    function = 'hll_cardinality'
    output_field = IntegerField()


class UnionAgg(Aggregate):
    function = 'hll_union_agg'
    output_field = HllField()


class UnionAggCardinality(Aggregate):
    """
    I haven't found a way to combine function inside function in django.
    So, I've written function to get aggregate cardinality with one call
    """
    function = 'hll_union_agg'
    template = 'hll_cardinality(%(function)s(%(expressions)s))'
    output_field = IntegerField()


class CardinalitySum(Aggregate):
    """
    I haven't found a way to combine function inside function in django.
    So, I've written function to get sum cardinality with one call
    """
    function = 'hll_cardinality'
    template = 'SUM(%(function)s(%(expressions)s))'
    output_field = IntegerField()



from collections import Iterable
from copy import deepcopy
from typing import Any

import six
from abc import abstractmethod, ABCMeta
from django.db.models.expressions import CombinedExpression, F, Func, Value


class HllJoinMixin:
    CONCAT = '||'

    def __or__(self, other):  # type: (Any) -> HllCombinedExpression
        # Functions, field references and other HllValues shouldn't be parsed
        if not isinstance(other, (F, HllValue, Func)):
            other = HllDataValue.parse_data(other)
        else:
            other = deepcopy(other)

        if isinstance(other, HllDataValue):
            other.added_to_hll_set()

        return HllCombinedExpression(self, self.CONCAT, other)


class HllCombinedExpression(HllJoinMixin, CombinedExpression):
    pass


class HllFromHex(six.with_metaclass(ABCMeta, Func)):
    """
    Constructs hll that can be saved from binary data (or it's psycopg representation)
    """
    def __init__(self, data, *args, **extra):
        db_type = extra.pop('db_type', 'hll')

        # Psycopg2 returns Binary results as hex string, prefixed by \x but requires bytes for saving.
        if isinstance(data, six.string_types) and data.startswith(r'\x'):
            data = bytearray.fromhex(data[2:])
        elif isinstance(data, bytes):
            pass
        else:
            raise ValueError('data should be bytes instance or string starting with \\x')

        self.template = extra.get('template', '%(expressions)s::{}'.format(db_type))

        super(HllFromHex, self).__init__(Value(data), *args, **extra)


class HllValue(six.with_metaclass(ABCMeta, HllJoinMixin, Func)):
    pass


class HllEmpty(HllValue):
    function = 'hll_empty'

    def __init__(self, **extra):
        super(HllEmpty, self).__init__(**extra)


class HllDataValue(HllValue):
    # This value is used to form real template and can be redeclared in descendants
    base_template = '%(function)s(%(expressions)s)'

    def __init__(self, data, *args, **extra):
        if not self.check(data):
            raise ValueError('Data is not supported by %s' % self.__class__.__name__)

        # hll_empty() is added in order to init set, if it's create or bulk operation
        if 'template' not in extra:
            extra['template'] = 'hll_empty() || %s' % self.base_template

        super(HllDataValue, self).__init__(Value(data), *args, **extra)

    def added_to_hll_set(self):  # type: () -> None
        # Remove hll_empty() prefix from value, it will be added by set
        self.extra['template'] = self.base_template

    @classmethod
    @abstractmethod
    def check(cls, data):  # type: (Any) -> bool
        """
        Checks if this value type can be used for given data
        :param data: Data to check
        :return: True, if this value can be used
        """
        raise NotImplementedError()

    @classmethod
    def parse_data(cls, data):  # type: (Any) -> HllDataValue
        # !!! Class order is important here!!! Can't use get_subclasses
        for klass in (HllBoolean, HllSmallInt, HllInteger, HllBigint, HllByteA, HllText, HllSet, HllAny):
            if klass.check(data):
                return klass(data)

        raise ValueError('No appropriate class found for value of type: %s' % str(type(data)))


class HllPrimitiveValue(six.with_metaclass(ABCMeta, HllDataValue)):
    def __init__(self, data, **extra):  # type: (Any, **dict) -> None
        """
        :param data: Data to build value from
        :param hash_seed: Optional hash seed. See https://github.com/citusdata/postgresql-hll#the-importance-of-hashing
        """
        extra['db_type'] = self.db_type

        hash_seed = extra.pop('hash_seed', None)
        if hash_seed is not None:
            super(HllPrimitiveValue, self).__init__(data, hash_seed, **extra)
        else:
            super(HllPrimitiveValue, self).__init__(data, **extra)

    @property
    def function(self):
        return 'hll_hash_%s' % self.db_type


class HllBoolean(HllPrimitiveValue):
    db_type = 'boolean'

    @classmethod
    def check(cls, data):
        return type(data) is bool


class HllIntegerValue(HllPrimitiveValue):
    # Abstract class property
    value_range = None
    base_template = '%(function)s(%(expressions)s::%(db_type)s)'

    @classmethod
    def check(cls, data):
        return cls.value_range and type(data) is int and cls.value_range[0] <= data <= cls.value_range[1]


class HllSmallInt(HllIntegerValue):
    db_type = 'smallint'
    value_range = (-32768, 32767)


class HllInteger(HllIntegerValue):
    db_type = 'integer'
    value_range = (-2147483648, 2147483647)


class HllBigint(HllIntegerValue):
    db_type = 'bigint'
    value_range = (-9223372036854775808, 9223372036854775807)


class HllByteA(HllPrimitiveValue):
    db_type = 'bytea'

    @classmethod
    def check(cls, data):
        return isinstance(data, bytes)


class HllText(HllPrimitiveValue):
    db_type = 'text'

    @classmethod
    def check(cls, data):
        return isinstance(data, six.string_types)


class HllAny(HllPrimitiveValue):
    db_type = 'any'

    @classmethod
    def check(cls, data):
        return True


class HllSet(HllValue):
    """
    Aggregate of HllValue objects
    """
    def __init__(self, *args, **extra):
        if args:
            if not self.check(args[0]):
                raise ValueError('Data is not supported by %s' % self.__class__.__name__)

            if isinstance(args[0], HllValue):
                self.data = (args[0],)

                if isinstance(args[0], HllDataValue):
                    args[0].added_to_hll_set()
            else:
                self.data = tuple(self._parse_item(item) for item in args[0])

            args = args[1:]
        else:
            self.data = tuple()

        super(HllSet, self).__init__(*args, **extra)

    @classmethod
    def _parse_item(cls, item):  # type: (Any) -> HllValue
        """
        Parses single item, checking if it can be added to hll and formatting it
        :param item: data to parse
        :return: HllDataValue instances
        """
        if isinstance(item, HllPrimitiveValue):
            pass
        elif isinstance(item, HllValue):
            raise ValueError("Only HllPrimitiveValue instances can be added to HllSet, not %s"
                             % item.__class__.__name__)
        else:
            item = HllDataValue.parse_data(item)

        item.added_to_hll_set()

        return item

    @classmethod
    def check(cls, data):
        return isinstance(data, (Iterable, HllValue))

    def as_sql(self, compiler, connection, function=None, template=None):
        sql, params = HllEmpty().as_sql(compiler, connection)

        for item in self.data:
            item_sql, item_params = item.as_sql(compiler, connection)
            sql += ' || %s' % item_sql
            params.extend(item_params)

        return sql, params




"""
This file contains a field to use in django models
"""
from django.db.models import BinaryField, IntegerField
from django.db.models.lookups import Transform

import six

from .values import HllEmpty, HllFromHex

__all__ = ['HllField']


class HllField(BinaryField):
    """
    A field, wrapping hll
    """
    description = "Postgres HyperLogLog"
    empty_values = [None, b'', HllEmpty()]

    # Default values were taken from https://github.com/citusdata/postgresql-hll#defaults
    custom_params = {
        'log2m': 11,
        'regwidth': 5,
        'expthresh': -1,
        'sparseon': 1
    }

    def __init__(self, *args, **kwargs):
        self._log2m = kwargs.pop('log2m', self.custom_params['log2m'])
        self._regwidth = kwargs.pop('regwidth', self.custom_params['regwidth'])
        self._expthresh = kwargs.pop('expthresh', self.custom_params['expthresh'])
        self._sparseon = kwargs.pop('sparseon', self.custom_params['sparseon'])

        super(HllField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(HllField, self).deconstruct()

        # Only include kwarg if it's not the default
        for param_name, default in self.custom_params.items():
            if getattr(self, '_%s' % param_name) != default:
                kwargs[param_name] = getattr(self, '_%s' % param_name)

        return name, path, args, kwargs

    def db_type(self, connection):
        return 'hll(%d, %d, %d, %d)' % (self._log2m, self._regwidth, self._expthresh, self._sparseon)

    def rel_db_type(self, connection):
        return 'hll'

    def get_internal_type(self):
        return self.__class__.__name__

    def get_db_prep_value(self, value, connection, prepared=False):
        # Psycopg2 returns Binary results as hex string, prefixed by \x
        # BinaryField requires bytes to be saved
        # But none of these can be converted to HLL by postgres directly
        if isinstance(value, bytes) or isinstance(value, six.string_types) and value.startswith(r'\x'):
            return HllFromHex(value, db_type=self.db_type(connection))
        else:
            return super(HllField, self).get_db_prep_value(value, connection, prepared=prepared)

    def get_default(self):
        if self.has_default() and not callable(self.default):
            return self.default
        default = super(HllField, self).get_default()
        if default == '':
            return HllEmpty()
        return default


@HllField.register_lookup
class CardinalityTransform(Transform):
    lookup_name = 'cardinality'
    output_field = IntegerField()

    def as_sql(self, compiler, connection, function=None, template=None):
        lhs, params = compiler.compile(self.lhs)
        return 'hll_cardinality(%s)' % lhs, params





from typing import TypeVar, Set

T = TypeVar('T')


def get_subclasses(cls, recursive=False):  # type: (T, bool) -> Set[T]
    """
    Gets all subclasses of given class
    Attention!!! Classes would be found only if they were imported before using this function
    :param cls: Class to get subcalsses
    :param recursive: If flag is set, returns subclasses of subclasses and so on too
    :return: A list of subclasses
    """
    subclasses = set(cls.__subclasses__())

    if recursive:
        for subcls in subclasses.copy():
            subclasses.update(get_subclasses(subcls, recursive=True))

    return subclasses





def django_pg_bulk_update_available():  # type: () -> bool
    """
    Tests if django-pb-bulk-update library is installed
    :return: Boolean
    """
    try:
        import django_pg_bulk_update
        return True
    except ImportError:
        return False
